# -*- coding: utf-8 -*-
#
# Copyright (c) 2012-2013, CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import SocketServer
import re

from actor import LIP

#: MLLP encoding chars
SB = "\x0b"
EB = "\x1c"
CR = "\x0d"

class MLLProtocol(object):
    """
    MLLProtocol class

    It just check if an incoming stream of bytes matches the following regexp:

    \x0b{string_containing_carriage_returns}\x1c\x0d
    """
    validator = re.compile(SB + "(([^\r]+\r){1,})" + EB + CR)

    @staticmethod
    def get_message(line):
        message = None
        matched = MLLProtocol.validator.match(line)
        if matched is not None:
            message = matched.groups()[0]
        return message

class MLLPServer(SocketServer.StreamRequestHandler):
    """
    Simplistic implementation of a TCP server implementing the MLLP protocol

    HL7 messages are encoded between bytes \x0b and \x1c\x0d
    """

    def handle(self):
        line = ''
        while True:
            char = self.rfile.read(1)
            if not char:
                print 'client disconnected'
                break
            line += char
            # check if incoming buffer contains a HL7 message
            message = MLLProtocol.get_message(line)
            if message is not None:
                # delegate message response to the LIP actor
                response = LIP.reply(message)
                # encode the response
                self.wfile.write(SB + response + CR + EB + CR)
                line = ''

if __name__ == "__main__":
    HOST, PORT = "localhost", 6000

    server = SocketServer.TCPServer((HOST, PORT), MLLPServer)
    server.serve_forever()


# -*- coding: utf-8 -*-
#
# Copyright (c) 2012-2013, CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""
HL7apy - helper functions
"""

import os
import sys
import collections
import importlib

from hl7apy.exceptions import UnsupportedVersion, InvalidEncodingChars, UnknownValidationLevel
from hl7apy.consts import DEFAULT_ENCODING_CHARS, DEFAULT_VERSION, VALIDATION_LEVEL

_DEFAULT_ENCODING_CHARS = DEFAULT_ENCODING_CHARS
_DEFAULT_VERSION = DEFAULT_VERSION
_DEFAULT_VALIDATION_LEVEL = VALIDATION_LEVEL.QUIET

def check_encoding_chars(encoding_chars):
    """
    Validate the given encoding chars

    :type encoding_chars: ``dict``
    :param encoding_chars: the encoding chars (see :func:`hl7apy.set_default_encoding_chars`)
    :raises: :exc:`hl7apy.exceptions.InvalidEncodingChars` if the given encoding chars are not valid
    """
    if not isinstance(encoding_chars, collections.MutableMapping):
        raise InvalidEncodingChars
    required = {'FIELD', 'COMPONENT', 'SUBCOMPONENT', 'REPETITION', 'ESCAPE'}
    missing = required - set(encoding_chars.keys())
    if missing:
        raise InvalidEncodingChars('Missing required encoding chars')

    values = [ v for k, v in encoding_chars.items() if k in required ]
    if len(values) > len(set(values)):
        raise InvalidEncodingChars('Found duplicate encoding chars')

def check_validation_level(validation_level):
    """
    Validate the given validation level

    :type validation_level: ``int``
    :param validation_level: validation level (see :class:`hl7apy.consts.VALIDATION_LEVEL`)
    :raises: :exc:`hl7apy.exceptions.UnknownValidationLevel` if the given validation level is unsupported
    """
    if validation_level not in (VALIDATION_LEVEL.QUIET, VALIDATION_LEVEL.STRICT):
        raise UnknownValidationLevel

def check_version(version):
    """
    Validate the given version number

    :type version: ``basestring``
    :param version: the version to validate (e.g. ``2.6``)
    :raises: :class:`hl7apy.exceptions.UnsupportedVersion` if the given version is unsupported
    """
    if version not in SUPPORTED_LIBRARIES:
        raise UnsupportedVersion(version)

def get_default_encoding_chars():
    """
    Get the default encoding chars

    :rtype: ``dict``
    :returns: the encoding chars (see :func:`hl7apy.set_default_encoding_chars`)

    >>> print get_default_encoding_chars()['FIELD']
    |
    """
    return _DEFAULT_ENCODING_CHARS

def get_default_version():
    """
    Get the default version

    :rtype: ``basestring``
    :returns: the default version

    >>> print get_default_version()
    2.5
    """
    return _DEFAULT_VERSION

def get_default_validation_level():
    """
    Get the default validation level

    :rtype: ``basestring``
    :returns: the default validation level

    >>> print get_default_validation_level()
    2
    """
    return _DEFAULT_VALIDATION_LEVEL

def set_default_validation_level(validation_level):
    """
    Set the given validation level as default

    :type validation_level: ``int``
    :param validation_level: validation level (see :class:`hl7apy.consts.VALIDATION_LEVEL`)
    :raises: :exc:`hl7apy.exceptions.UnknownValidationLevel` if the given validation level is unsupported

    >>> set_default_validation_level(3)
    Traceback (most recent call last):
        ...
    UnknownValidationLevel
    >>> set_default_validation_level(VALIDATION_LEVEL.QUIET)
    >>> print get_default_validation_level()
    2
    """
    check_validation_level(validation_level)

    global _DEFAULT_VALIDATION_LEVEL

    _DEFAULT_VALIDATION_LEVEL = validation_level

def set_default_version(version):
    """
    Set the given version as default

    :type version: ``basestring``
    :param version: the new default version (e.g. ``2.6``)
    :raises: :class:`hl7apy.exceptions.UnsupportedVersion` if the given version is unsupported

    >>> set_default_version('22')
    Traceback (most recent call last):
        ...
    UnsupportedVersion: The version 22 is not supported
    >>> set_default_version('2.3')
    >>> print get_default_version()
    2.3
    """
    check_version(version)

    global _DEFAULT_VERSION

    _DEFAULT_VERSION = version

def set_default_encoding_chars(encoding_chars):
    """
    Set the given encoding chars as default

    :type encoding_chars: ``dict``
    :param encoding_chars: the new encoding chars
    :raises: :class:`hl7apy.exceptions.InvalidEncodingChars` if the given encoding chars are not valid

    The *encoding_chars* dictionary should contain the following keys:

    +--------------+-----------------+
    |Key           |Default          |
    +==============+=================+
    |GROUP         |``\\r``           |
    +--------------+-----------------+
    |SEGMENT       |``\\r``           |
    +--------------+-----------------+
    |FIELD         |``|``            |
    +--------------+-----------------+
    |COMPONENT     |``^``            |
    +--------------+-----------------+
    |SUBCOMPONENT  |``&``            |
    +--------------+-----------------+
    |REPETITION    |``~``            |
    +--------------+-----------------+
    |ESCAPE        |``\\``            |
    +--------------+-----------------+

    >>> set_default_encoding_chars({'FIELD': '!'})
    Traceback (most recent call last):
        ...
    InvalidEncodingChars: Invalid encoding chars
    >>> set_default_encoding_chars({'FIELD': '!', 'COMPONENT': 'C', 'SUBCOMPONENT': 'S', 'REPETITION': 'R', 'ESCAPE': '\\\\'})
    >>> print get_default_encoding_chars()['FIELD']
    !
    """
    check_encoding_chars(encoding_chars)

    encoding_chars.update({'GROUP' : '\r', 'SEGMENT' : '\r'})

    global _DEFAULT_ENCODING_CHARS

    _DEFAULT_ENCODING_CHARS = encoding_chars

def load_library(version):
    """
    Load the correct module according to the version

    :type version: ``basestring``
    :param version: the version of the library to be loaded (e.g. '2.6')
    :rtype: module object
    """
    check_version(version)
    module_name = SUPPORTED_LIBRARIES[version]
    lib = sys.modules.get(module_name)
    if lib is None:
        lib = importlib.import_module(module_name)
    return lib

def load_reference(name, element_type, version):
    """
    Look for an element of the given type, name and version and return its reference structure

    :type element_type: ``basestring``
    :param element_type: the element type to look for (e.g. 'Segment')
    :type name: ``basestring``
    :param name: the element name to look for (e.g. 'MSH')
    :type version: ``basestring``
    :param version: the version of the library where to search the element (e.g. '2.6')
    :rtype: ``dict``
    :return: a dictionary describing the element structure
    :raise: ``KeyError`` if the element has not been found

    The returned dictionary will contain the following keys:

    +--------------+--------------------------------------------+
    |Key           |Value                                       |
    +==============+============================================+
    |cls           |an :class:`hl7apy.core.Element` subclass    |
    +--------------+--------------------------------------------+
    |name          |the Element name (e.g. PID)                 |
    +--------------+--------------------------------------------+
    |ref           |a tuple of one of the following format:     |
    |              |                                            |
    |              |('leaf', <datatype>, <longName>, <table>)   |
    |              |('sequence', (<child>, (<min>, <max>), ...))|
    +--------------+--------------------------------------------+

    >>> load_reference('UNKNOWN', 'Segment', '2.5')
    Traceback (most recent call last):
    ...
    KeyError: 'UNKNOWN'
    >>> r = load_reference('ADT_A01', 'Message', '2.5')
    >>> print r[0]
    sequence
    >>> r = load_reference('MSH_3', 'Field', '2.5')
    >>> print r[0]
    leaf
    """
    lib = load_library(version)
    ref = lib.get(name, element_type)
    return ref

def find_reference(name, element_types, version):
    """
    Look for an element of the given name and version into the given types and return its reference structure

    :type name: ``basestring``
    :param name: the element name to look for (e.g. 'MSH')
    :type types: ``list`` or ``tuple``
    :param types: the element classes where to look for the element (e.g. (Group, Segment))
    :type version: ``basestring``
    :param version: the version of the library where to search the element (e.g. '2.6')
    :rtype: ``dict``
    :return: a dictionary describing the element structure
    :raise: :class:`hl7apy.exceptions.ChildNotFound` if the element has not been found

    >>> from hl7apy.core import Message, Segment
    >>> find_reference('UNKNOWN', (Segment, ), '2.5')
    Traceback (most recent call last):
    ...
    ChildNotFound: No child named UNKNOWN
    >>> find_reference('ADT_A01', (Segment,),  '2.5')
    Traceback (most recent call last):
    ...
    ChildNotFound: No child named ADT_A01
    >>> r = find_reference('ADT_A01', (Message,),  '2.5')
    >>> print r['name'], r['cls']
    ADT_A01 <class 'hl7apy.core.Message'>
    """
    lib = load_library(version)
    ref = lib.find(name, element_types)
    return ref

def _discover_libraries():
    current_dir = os.path.dirname(__file__)
    return {o[1:].replace("_", "."): "hl7apy.{}".format(o)
            for o in os.listdir(current_dir) if o.startswith("v2_")}

SUPPORTED_LIBRARIES = _discover_libraries()

if __name__ == '__main__':

    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2012-2013, CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
.. warning::

    The HL7 versions can have different implementation of base datatypes; for example the ST base datatype
    of HL7 v2.6 is different from the v2.5 one. This module contains reference classes for all base datatypes
    but you should not import them directly from here. If you need an implementation for a particular version use the
    :func:`get_base_datatypes` function from a specific version's module
    For example if you're using version 2.4 and you need an `FT` base datatype do the following:

    >>> from hl7apy.v2_4 import get_base_datatypes
    >>> FT = get_base_datatypes()['FT']
    >>> f = FT('some useful information')

"""

import re
import numbers
from datetime import datetime

from decimal import Decimal

from hl7apy import get_default_encoding_chars, get_default_validation_level
from hl7apy.exceptions import MaxLengthReached, InvalidHighlightRange, InvalidDateFormat, InvalidDateOffset
from hl7apy.validation import Validator


class BaseDataType(object):
    """
    Generic datatype base class. It handles the value of the data type and its
    maximum length. It is meant to be extended and it should not be used directly

    :param value: the value of the data type

    :type max_length: ``int``
    :param max_length: The maximum length of the value. Default to None

    :type validation_level: ``int``
    :param validation_level: It must be a value from class :class:`hl7apy.consts.VALIDATION_LEVEL`
        If it is :attr:`STRICT` it checks that :attr:`value` doesn't exceed the attr:`max_length`

    :raise: :exc:`hl7apy.exceptions.MaxLengthReached` When the :attr:`value`'s length is greater than the
     :attr:`max_length`. Only if :attr:`validation_level` is :attr:`STRICT`
    """
    def __init__(self, value, max_length=None, validation_level=None):
        if validation_level is None:
            validation_level = get_default_validation_level()
        self.validation_level = validation_level
        self.max_length = max_length
        if Validator.is_strict(self.validation_level):
            if self.max_length is not None and \
                len('{0}'.format(value)) > self.max_length:
                raise MaxLengthReached(value, self.max_length)
        self.value = value

    def to_er7(self, encoding_chars=None):
        """
        Encode to ER7 format
        """
        return '{0}'.format(self.value or '')

    @property
    def classname(self):
        """
        The name of the class
        """
        return self.__class__.__name__


class TextualDataType(BaseDataType):
    """
    Base class for textual data types.
    It is meant to be extended and it should not be used directly

    :type value: ``basestring``
    :param value: the value of the data type

    :type max_length: ``int``
    :param max_length: the max length of the value (default to 32)

    :type highlights: ``tuple``, ``list``
    :param highlights: a list of ranges indicating the part
        of the value to be highlighted. e.g. ((0,5), (6,7))
        The highlights cannot overlap, if they do an HL7Exception will be
        thrown when to_er7 method is called

    :type validation_level: ``int``
    :param validation_level: It has the same meaning as in :class:`BaseDatatype`

    :raise: :exc:`hl7apy.exceptions.MaxLengthReached` When the `value`'s length is greater than `max_length`
    """
    def __init__(self, value, max_length=32, highlights=None,
                 validation_level=None):
        self.highlights = highlights
        super(TextualDataType, self).__init__(value, max_length,
                                              validation_level)

    def to_er7(self, encoding_chars=None):
        if encoding_chars is None:
            encoding_chars = get_default_encoding_chars()
        return self._escape_value(self.value, encoding_chars)

    def _escape_value(self, value, encoding_chars=None):
        escape_char = encoding_chars['ESCAPE']
        translations = ((encoding_chars['FIELD'], '{esc}F{esc}'.format(esc=escape_char)),
                       (encoding_chars['COMPONENT'], '{esc}S{esc}'.format(esc=escape_char)),
                       (encoding_chars['SUBCOMPONENT'], '{esc}T{esc}'.format(esc=escape_char)),
                       (encoding_chars['REPETITION'], '{esc}R{esc}'.format(esc=escape_char)))

        # Inserts the highlights escape sequences
        if self.highlights is not None:
            def _sort_highlights(x, y):
                if x[0] < y[0]:
                    if y[0] < x[1]: #overlapping ranges
                        raise InvalidHighlightRange(x, y)
                    else:
                        return -1
                elif x[0] > y[0]:
                    if x[0] < y[1]: #overlapping ranges
                        raise InvalidHighlightRange(x, y)
                    else:
                        return 1
                else:
                    raise InvalidHighlightRange(x, y)

            self.highlights = sorted(self.highlights, cmp=_sort_highlights)
            words = list(value)
            offset = 0
            for hl in self.highlights:
                if hl[0] > hl[1]:
                    raise InvalidHighlightRange(hl[0], hl[1])
                words.insert(hl[0] + offset, '{esc}H{esc}'.format(esc=escape_char))
                words.insert(hl[1] + offset + 1, '{esc}N{esc}'.format(esc=escape_char))
                offset += 2
            value = ''.join(words)

        # Escapes encoding_chars
        for char, esc_seq in translations:
            value = value.replace(char, esc_seq)
        # Escapes the escape_char. If it is found in other escape sequences it is not escaped.
        # For example if the escape char is / and we find /H/ the escape chars are not re-escaped, otherwise it would become
        # /E/H/E/ which is not the result wanted.
        # Thus the regex search for escape chars not followed and not preceeded by one of the litteral composing an escape sequence
        # We use lambda because otherwise the backslash sequence in the string is processed (look for re.sub in python doc) and we don't want this
        value = re.sub('(?<!%s[HNFSTRE])%s(?![HNFSTRE]%s)' % tuple(3*[re.escape(escape_char)]),
                        lambda x: '{esc}E{esc}'.format(esc=escape_char), value)

        return value


class NumericDataType(BaseDataType):
    """
    Base class for numeric data types.
    It is meant to be extended and it should not be used directly

    :param value: the value of the data type. Default is ``None``

    :type max_length: int
    :param max_length: The maximum number of digit in the value. Default is 16

    :type validation_level: ``int``
    :param validation_level: It has the same meaning as in :class:`hl7apy.base_datatypes.BaseDataType`

    :raise: :exc:`hl7apy.exceptions.MaxLengthReached` When the `value`'s length is greater than `max_length`

    """
    def __init__(self, value=None, max_length=16,
                 validation_level=None):
        super(NumericDataType, self).__init__(value, max_length,
                                              validation_level)


class DateTimeDataType(BaseDataType):
    """
    Base class for datetime data types.
    It is meant to be extended and it should not be used directly.
    Children classes should at least override the :attr:`allowed_formats` tuple

    :type value datetime: datetime.datetime
    :param value: a `datetime` date object. Default is ``None``

    :type format: str
    :param format: the format that will be used converting the object to string.
        It must be an item of the :attr:`allowed_formats` tuple

    :raise: :exc:``hl7apy.exceptions.InvalidDateFormat` if the ``format`` is not in the :attr:`allowed_formats` member
    """

    allowed_formats = ()

    def __init__(self, value=None, format=''):
        if format not in self.allowed_formats:
            raise InvalidDateFormat(format)
        self.value = value
        self.format = format

    def to_er7(self, encoding_chars=None):
        return datetime.strftime(self.value, self.format)


class DT(DateTimeDataType):
    """
    Class for DT base datatype. It extends DatetimeDatatype and it represents a time value with
    year, month and day. Parameters are the same of the superclass.

    The :attr:`allowed_formats` tuple is ``('%Y', '%Y%m', '%Y%m%d')``
    """

    allowed_formats = ('%Y', '%Y%m', '%Y%m%d')

    def __init__(self, value=None, format='%Y%m%d'):
        super(DT, self).__init__(value, format)


class TM(DateTimeDataType):
    """
    Class for TM base datatype. It extends DateTimeDatatype and it represents a time value with
    hours, minutes, seconds and microseconds. Parameters are the same of the superclass plus ``offset``

    The :attr:`allowed_formats` tuple is ``('%H', '%H%M', '%H%M%S', '%H%M%S.%f')``.
    It needs also the ``offset`` parameter which represents the UTC offset

    :type offset: ``basestring``
    :param offset: the UTC offset. By default it is ''. It must be in the form ``'+/-HHMM'``
    """

    allowed_formats = ('%H', '%H%M', '%H%M%S', '%H%M%S.%f')

    def __init__(self, value=None, format='%H%M%S.%f', offset=''):
        super(TM, self).__init__(value, format)

        if offset and len(offset) != 5:
            raise InvalidDateOffset(offset)
        try:
            d = datetime.strptime(offset[1:], '%H%M')
            if d.hour > 12:
                raise ValueError
        except ValueError:
            if offset:
                raise InvalidDateOffset(offset)

        if offset and offset[0] not in ('+', '-'):
            raise InvalidDateOffset(offset)

        self.offset = offset

    def to_er7(self, encoding_chars=None):
        return '{0}{1}'.format(super(TM, self).to_er7(), self.offset)


class DTM(TM):
    """
    Class for TM base datatype. It extends TM and it represents classes DT and DTM combined.
    Thus it represents year, month, day, hours, minutes, seconds and microseconds.
    Parameters are the same of the superclass.

    The :attr:`allowed_formats` tuple is
    ``('%Y', '%Y%m', '%Y%m%d', '%Y%m%d%H', '%Y%m%d%H%M', '%Y%m%d%H%M%S', '%Y%m%d%H%M%S.%f')``
    """

    allowed_formats = ('%Y', '%Y%m', '%Y%m%d', '%Y%m%d%H', '%Y%m%d%H%M',
                       '%Y%m%d%H%M%S', '%Y%m%d%H%M%S.%f')

    def __init__(self, value=None, format='%Y%m%d%H%M%S.%f', offset=''):
        super(DTM, self).__init__(value, format, offset)


class ST(TextualDataType):
    """
    Class for ST datatype. It extends :class:`hl7apy.base_datatypes.TextualDatatype` and the parameters are the same of the superclass

    :attr:`max_length` is 199
    """
    def __init__(self, value, highlights=None,
                 validation_level=None):
        super(ST, self).__init__(value, 199, highlights, validation_level)


class FT(TextualDataType):
    """
    Class for FT datatype. It extends :class:`hl7apy.base_datatypes.TextualDataType` and the parameters are the same of the superclass

    :attr:`max_length` is 65536
    """
    def __init__(self, value, highlights=None,
                 validation_level=None):
        super(FT, self).__init__(value, 65536, highlights, validation_level)


class ID(TextualDataType):
    """
    Class for ID datatype. It extends :class:`hl7apy.base_datatypes.TextualDataType` and the parameters are the same of the superclass

    :attr:`max_length` None
    """
    def __init__(self, value, highlights=None,
                 validation_level=None):
        #max_length is None bacause it depends from the HL7 table
        super(ID, self).__init__(value, None, highlights, validation_level)
        #TODO: check for tables of allowed values: are we strict or not?


class IS(TextualDataType):
    """
    Class for IS datatype. It extends :class:`hl7apy.base_datatypes.TextualDataType` and the parameters are the same of the superclass

    :attr:`max_length` is 20
    """
    def __init__(self, value, highlights=None,
                 validation_level=None):
        super(IS, self).__init__(value, 20, highlights, validation_level)
        #TODO: check for tables of allowed values (also defined on site): are we strict or not?


class TX(TextualDataType):
    """
    Class for TX datatype. It extends :class:`hl7apy.base_datatypes.TextualDataType` and the parameters are the same of the superclass

    :attr:`max_length` is 65536
    """
    def __init__(self, value, highlights=None,
                 validation_level=None):
        super(TX, self).__init__(value, 65536, highlights, validation_level)


class GTS(TextualDataType):
    """
    Class for GTS datatype. It extends :class:`hl7apy.base_datatypes.TextualDataType` and the parameters are the same of the superclass

    :attr:`max_length` is 199
    """
    def __init__(self, value, highlights=None,
                 validation_level=None):
        super(GTS, self).__init__(value, 199, highlights, validation_level)


class NM(NumericDataType):
    """
    Class for NM datatype. It extends :class:`hl7apy.base_datatypes.NumericDatatype` and the parameters are the same of the superclass

    :attr:`max_length` is 16.

    The type of ``value`` must be :class:`decimal.Decimal` or :class:`numbers.Real`

    :raise: :exc:`ValueError` raised when the value is not of one of the correct type
    """

    def __init__(self, value=None,
                 validation_level=None):
        if value is not None and isinstance(value, numbers.Real):
            value = Decimal(value)
        elif value is not None and not isinstance(value, Decimal):
            raise ValueError('Invalid value for a NM data')
        super(NM, self).__init__(value, 16, validation_level)


class SI(NumericDataType):
    """
    Class for NM datatype. It extends NumericDatatype and the parameters are the same of the superclass

    :attr:`max_length` is 4.

    The type of ``value`` must be `int` or :class:`numbers.Integral`

    :raise: :exc:`ValueError` raised when the value is not of one of the correct type
    """
    def __init__(self, value=None,
                 validation_level=None):
        if value and not isinstance(value, numbers.Integral):
            raise ValueError('Invalid value for a SI data')

        super(SI, self).__init__(value, 4, validation_level)


class TN(TextualDataType):
    """
    Class for TN datatype. It extends TextualDatatype and the parameters are the same of the superclass

    :attr:`max_length` is 199.

    The type of ``value`` must be `basestring` and should match the format
    [NN] [(999)]999-9999[X99999][B99999][C any text]

    :raise: :exc:`ValueError` raised when the value does not match the expected format
    """
    def __init__(self, value, validation_level=None):

        regexp = "(\d\d\s)?(\(\d+\))?(\d+-?\d+)(X\d+)?(B\d+)?(C.+)?"
        if not re.match(regexp, value):
            raise ValueError('Invalid value for TN data')

        super(TN, self).__init__(value, 199, None, validation_level)




# -*- coding: utf-8 -*-
#
# Copyright (c) 2012-2013, CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import importlib

from messages import MESSAGES
from segments import SEGMENTS
from fields import FIELDS
from datatypes import DATATYPES
from groups import GROUPS

from hl7apy.exceptions import ChildNotFound

ELEMENTS = {'Message': MESSAGES, 'Segment': SEGMENTS, 'Field': FIELDS,
            'Component': DATATYPES, 'Group': GROUPS, 'SubComponent': DATATYPES}

def get(name, element_type):
    return ELEMENTS[element_type][name]

def find(name, where):
    for cls in where:
        try:
            return {'ref': get(name, cls.__name__), 'name': name, 'cls': cls}
        except:
            pass
    raise ChildNotFound(name)

def is_base_datatype(datatype):
    return datatype in BASE_DATATYPES

def get_base_datatypes():
    return BASE_DATATYPES

def _load_base_datatypes():
    base_datatypes = ('ST', 'ID', 'DT', 'FT', 'IS', 'NM', 'SI', 'TM', 'TX', 'TN')
    module = importlib.import_module("hl7apy.base_datatypes")
    datatypes = {}
    for cls in base_datatypes:
        cls = getattr(module, cls)
        datatypes[cls.__name__] = cls
    return datatypes

BASE_DATATYPES = _load_base_datatypes()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2012-2013, CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""
This module contains factory functions for hl7apy base data types.
The functions get the value of the data type as string and return the correct object
"""

import re
from datetime import datetime
from decimal import Decimal, InvalidOperation
from types import FunctionType

from hl7apy import load_library, get_default_validation_level, get_default_version
from hl7apy.exceptions import InvalidDataType

def datatype_factory(datatype, value, version=None, validation_level=None):
    """
    Factory function for both base and complex datatypes. It generates the correct object according
    to the datatype in input.
    It should be noted that if you use the factory it is not possible to specify
    some parameters for the datatype (e.g. the format for datetime base datatypes)
    If the value is not valid for the datatype specified if the ``validation_level`` is
    :attr:`hl7apy.consts.VALIDATION_LEVEL.QUIET` it generates an :class:`hl7apy.base_datatypes.ST` object

    :type datatype: ``basestring``
    :param datatype: The datatype to be generated

    :param value: The value of the datatype

    :type version: ``basestring``
    :param version: A valid HL7 version. It must be one of :attr:`hl7apy.SUPPORTED_LIBRARIES`

    :type validation_level: ``int``
    :param validation_level: It must be a value from class :attr:`validation_level` :class:`VALIDATION_LEVEL`

    :rtype: The type specified in datatype

    :raises :exc:`hl7apy.exceptions.ValueError`: If the ``validation_level`` is :attr:`hl7apy.consts.VALIDATION_LEVEL.STRICT`
     and the value is not valid for the specified datatype

    :raises :exc:`hl7apy.exceptions.InvalidDatatype`: If the ``datatype`` specified is not valid for the given ``version``

    """

    from hl7apy.validation import Validator

    if validation_level is None:
        validation_level = get_default_validation_level()

    if version is None:
        version = get_default_version()

    lib = load_library(version)

    base_datatypes = lib.get_base_datatypes()

    factories = base_datatypes.copy()

    if 'DT' in factories:
        factories['DT'] = date_factory
    if 'TM' in factories:
        factories['TM'] = timestamp_factory
    if 'DTM' in factories:
        factories['DTM'] = datetime_factory
    if 'NM' in factories:
        factories['NM'] = numeric_factory
    if 'SI' in factories:
        factories['SI'] = sequence_id_factory

    try:
        factory = factories[datatype]
        if isinstance(factory, FunctionType):
            return factory(value, base_datatypes[datatype])
        return factory(value)
    except KeyError:
        raise InvalidDataType(datatype)
    except ValueError, e:
        if Validator.is_strict(validation_level):
            raise e
        #TODO: Do we really want this? In that case the parent's datatype must be changed accordingly
        return factories['ST'](value)

def date_factory(value, datatype_cls):
    """
    Creates a :class:`hl7apy.base_datatypes.DT` object

    The value in input must be a string parsable with :meth:`datetime.strptime`.
    The date format is chosen according to the length of the value as stated in this table:

    +-------+-----------+
    |Length |Format     |
    +=======+===========+
    |4      |``%Y``     |
    |       |           |
    +-------+-----------+
    |6      |``%Y%m``   |
    |       |           |
    +-------+-----------+
    |8      |``%Y%m%d`` |
    |       |           |
    +-------+-----------+

    Some examples that work are:

    >>> from hl7apy.base_datatypes import DT
    >>> date_factory("1974", DT) #doctest: +ELLIPSIS
    <hl7apy.base_datatypes.DT object at 0x...>
    >>> date_factory("198302", DT) #doctest: +ELLIPSIS
    <hl7apy.base_datatypes.DT object at 0x...>
    >>> date_factory("19880312", DT) #doctest: +ELLIPSIS
    <hl7apy.base_datatypes.DT object at 0x...>

    If the value does not match one of the valid format it raises :exc:`ValueError`

    :type value: ``basestring``
    :param value: the value to assign the date object

    :type datatype_cls: `class`
    :param value: the :class:`hl7apy.base_datatypes.DT` class to use. It has to be one implementation of the different version modules

    :rtype: :class:`hl7apy.base_datatypes.DT`
    """

    format = _get_date_format(value)
    dt_value = _datetime_obj_factory(value, format)
    return datatype_cls(dt_value, format)

def timestamp_factory(value, datatype_cls):
    """
    Creates a :class:`hl7apy.base_datatypes.TM` object

    The value in input must be a string parsable with :meth:`datetime.strptime`.
    It can also have an offset part specified with the format +/-HHMM.
    The offset can be added with all the allowed format
    The date format is chosen according to the length of the value as stated in this table:

    +-------+-----------------+
    |Length |Format           |
    +=======+=================+
    |2      |``%H``           |
    +-------+-----------------+
    |4      |``%H%M``         |
    +-------+-----------------+
    |6      |``%H%M%S``       |
    +-------+-----------------+
    |10-13  |``%H%M%S.%f``    |
    +-------+-----------------+

    Some examples that work are:


    >>> from hl7apy.base_datatypes import TM
    >>> timestamp_factory("12", TM) #doctest: +ELLIPSIS
    <hl7apy.base_datatypes.TM object at 0x...>
    >>> timestamp_factory("12+0300", TM) #doctest: +ELLIPSIS
    <hl7apy.base_datatypes.TM object at 0x...>
    >>> timestamp_factory("1204", TM) #doctest: +ELLIPSIS
    <hl7apy.base_datatypes.TM object at 0x...>
    >>> timestamp_factory("120434", TM) #doctest: +ELLIPSIS
    <hl7apy.base_datatypes.TM object at 0x...>
    >>> timestamp_factory("120434-0400", TM) #doctest: +ELLIPSIS
    <hl7apy.base_datatypes.TM object at 0x...>

    If the value does not match one of the valid format it raises :exc:ValueError`

    :type value: ``basestring``
    :param value: the value to assign the date object

    :type datatype_cls: `class`
    :param value: the :class:`hl7apy.base_datatypes.TM` class to use. It has to be one implementation of the different version modules

    :rtype: :class:`hl7apy.base_datatypes.TM`
    """

    value, offset = _split_offset(value)
    form = _get_timestamp_format(value)
    dt_value = _datetime_obj_factory(value, form)
    return datatype_cls(dt_value, form, offset)

def datetime_factory(value, datatype_cls):
    """
    Creates a :class:`hl7apy.base_datatypes.DTM` object

    The value in input must be a string parsable with :meth:`datetime.strptime`.
    It can also have an offset part specified with the format +HHMM -HHMM.
    The offset can be added with all the allowed format.
    The date format is chosen according to the length of the value as stated in this table:

    +-------+-----------------------+
    |Length |Format                 |
    +=======+=======================+
    |4      |``%Y``                 |
    +-------+-----------------------+
    |6      |``%Y%m``               |
    +-------+-----------------------+
    |8      |``%Y%m%d``             |
    +-------+-----------------------+
    |10     |``%Y%m%d%H``           |
    +-------+-----------------------+
    |12     |``%Y%m%d%H%M``         |
    +-------+-----------------------+
    |14     |``%Y%m%d%H%M%S``       |
    +-------+-----------------------+
    |18-21  |``%Y%m%d%H%M%S.%f``    |
    +-------+-----------------------+

    Some examples that work are:


    >>> from hl7apy.base_datatypes import DTM
    >>> datetime_factory("1924", DTM) #doctest: +ELLIPSIS
    <hl7apy.base_datatypes.DTM object at 0x...>
    >>> datetime_factory("1924+0300", DTM) #doctest: +ELLIPSIS
    <hl7apy.base_datatypes.DTM object at 0x...>
    >>> datetime_factory("19220430", DTM) #doctest: +ELLIPSIS
    <hl7apy.base_datatypes.DTM object at 0x...>
    >>> datetime_factory("19220430-0400", DTM) #doctest: +ELLIPSIS
    <hl7apy.base_datatypes.DTM object at 0x...>

    If the value does not match one of the valid format it raises :exc:`ValueError`

    :type value: ``basestring``
    :param value: the value to assign the date object

    :type datatype_cls: `class`
    :param value: the :class:`hl7apy.base_datatypes.DTM` class to use. It has to be one implementation of the different version modules

    :rtype: :class:`hl7apy.base_datatypes.DTM`
    """

    value, offset = _split_offset(value)
    date_format = _get_date_format(value[:8])
    try:
        timestamp_form = _get_timestamp_format(value[8:])
    except ValueError as e:
        if not value[8:]: #if it's empty
            timestamp_form = ''
        else:
            raise e

    form = '{0}{1}'.format(date_format, timestamp_form)
    dt_value = _datetime_obj_factory(value, form)
    return datatype_cls(dt_value, form, offset)

def numeric_factory(value, datatype_cls):
    """
    Creates a :class:`hl7apy.base_datatypes.NM` object

    The value in input can be a string representing a decimal number or a ``float``.
    (i.e. a string valid for :class:`decimal.Decimal()`).
    If it's not, a :exc:`hl7apy.exceptions.ValueError` is raised
    Also an empty string or ``None`` are allowed

    :type value: ``basestring`` or ``None``
    :param value: the value to assign the numeric object

    :type datatype_cls: :class:`hl7apy.base_datatypes.NM`
    :param value: the :class:`hl7apy.base_datatypes.NM` class to use. It has to be one implementation of the different version modules

    :rtype: :class:`hl7apy.base_datatypes.NM`
    """
    if not value:
        return datatype_cls()
    try:
        return datatype_cls(Decimal(value))
    except InvalidOperation:
        raise ValueError('{0} is not an HL7 valid NM value'.format(value))

def sequence_id_factory(value, datatype_cls):
    """
    Creates a :class:`hl7apy.base_datatypes.SI` object

    The value in input can be a string representing an integer number or an ``int``.
    (i.e. a string valid for ``int()``  ).
    If it's not, a :exc:`ValueError` is raised
    Also an empty string or ``None`` are allowed

    :type value: ``basestring`` or ``None``
    :param value: the value to assign the date object

    :type datatype_cls: `class`
    :param value: the SI class to use. It has to be loaded from one implementation of the different version modules

    :rtype: :class:`hl7apy.base_datatypes.SI`
    """
    if not value:
        return datatype_cls()
    try:
        return datatype_cls(int(value))
    except ValueError:
        raise ValueError('{0} is not an HL7 valid SI value'.format(value))

def _split_offset(value):
    offset = re.search('\d*((-|\+)(1[0-2]|0[0-9])([0-5][0-9]))$', value)
    if offset:
        offset = offset.groups()[0]
        return value.replace(offset, ''), offset
    return value, ''

def _get_date_format(value):
    if len(value) == 4:
        format = '%Y'
    elif len(value) == 6:
        format = '%Y%m'
    elif len(value) == 8:
        format = '%Y%m%d'
    else:
        raise ValueError('{0} is not an HL7 valid date value'.format(value))

    return format

def _get_timestamp_format(value):
    if len(value) == 2:
        format = '%H'
    elif len(value) == 4:
        format = '%H%M'
    elif len(value) == 6:
        format = '%H%M%S'
    elif 10 <= len(value) <= 13 and value[9] == '.':
        format = '%H%M%S.%f'
    else:
        raise ValueError('{0} is not an HL7 valid date value'.format(value))

    return format

def _datetime_obj_factory(value, format):
    try:
        dt_value = datetime.strptime(value, format)
    except ValueError:
        raise ValueError('{0} is not an HL7 valid date value'.format(value))
    return dt_value


if __name__ == '__main__':
    import doctest
    doctest.testmod()